@@ -23,22 +23,20 @@
 #ifndef __GST_I18N_LIB_H__
 #define __GST_I18N_LIB_H__
 
-#ifndef GST_VERSION
+#ifndef PACKAGE_NAME
 #error You must include config.h before including this header.
 #endif
 
-#ifdef ENABLE_NLS
-
 #include "gettext.h" /* included with gettext distribution and copied */
 
+#ifdef ENABLE_NLS
+
 /* we want to use shorthand _() for translating and N_() for marking */
 #define _(String) dgettext (GETTEXT_PACKAGE, String)
 #define N_(String) gettext_noop (String)
 /* FIXME: if we need it, we can add Q_ as well, like in glib */
 
 #else
-#undef GETTEXT_PACKAGE /* could still be defined in config.h */
-#define GETTEXT_PACKAGE NULL
 #define _(String) String
 #define N_(String) String
 
