@@ -24,8 +24,9 @@
  * @short_description: Asynchronous message bus subsystem
  * @see_also: #GstMessage, #GstElement
  *
- * The #GstBus is an object responsible for delivering #GstMessages in
- * a first-in first-out way from the streaming threads to the application.
+ * The #GstBus is an object responsible for delivering #GstMessage packets in
+ * a first-in first-out way from the streaming threads (see #GstTask) to the
+ * application.
  *
  * Since the application typically only wants to deal with delivery of these
  * messages from one thread, the GstBus will marshall the messages between
@@ -44,12 +45,12 @@
  *
  * The bus can be polled with the gst_bus_poll() method. This methods blocks
  * up to the specified timeout value until one of the specified messages types
- * is posted on the bus. The application can then _pop() the messages from the
- * bus to handle them.
+ * is posted on the bus. The application can then gst_bus_pop() the messages
+ * from the bus to handle them.
  * Alternatively the application can register an asynchronous bus function
  * using gst_bus_add_watch_full() or gst_bus_add_watch(). This function will
- * install a #GSource in the default glib main loop and will deliver messages 
- * a short while after they have been posted. Note that the main loop should 
+ * install a #GSource in the default glib main loop and will deliver messages
+ * a short while after they have been posted. Note that the main loop should
  * be running for the asynchronous callbacks.
  *
  * It is also possible to get messages from the bus without any thread
@@ -91,10 +92,6 @@ static void gst_bus_class_init (GstBusClass * klass);
 static void gst_bus_init (GstBus * bus);
 static void gst_bus_dispose (GObject * object);
 
-static void gst_bus_set_property (GObject * object, guint prop_id,
-    const GValue * value, GParamSpec * pspec);
-static void gst_bus_get_property (GObject * object, guint prop_id,
-    GValue * value, GParamSpec * pspec);
 static void gst_bus_set_main_context (GstBus * bus, GMainContext * ctx);
 
 static GstObjectClass *parent_class = NULL;
@@ -103,37 +100,12 @@ static guint gst_bus_signals[LAST_SIGNAL] = { 0 };
 struct _GstBusPrivate
 {
   guint num_sync_message_emitters;
-
   GCond *queue_cond;
-
   GSource *watch_id;
-
   GMainContext *main_context;
 };
 
-GType
-gst_bus_get_type (void)
-{
-  static GType bus_type = 0;
-
-  if (G_UNLIKELY (bus_type == 0)) {
-    static const GTypeInfo bus_info = {
-      sizeof (GstBusClass),
-      NULL,
-      NULL,
-      (GClassInitFunc) gst_bus_class_init,
-      NULL,
-      NULL,
-      sizeof (GstBus),
-      0,
-      (GInstanceInitFunc) gst_bus_init,
-      NULL
-    };
-
-    bus_type = g_type_register_static (GST_TYPE_OBJECT, "GstBus", &bus_info, 0);
-  }
-  return bus_type;
-}
+G_DEFINE_TYPE (GstBus, gst_bus, GST_TYPE_OBJECT);
 
 /* fixme: do something about this */
 static void
@@ -166,20 +138,11 @@ marshal_VOID__MINIOBJECT (GClosure * closure, GValue * return_value,
 static void
 gst_bus_class_init (GstBusClass * klass)
 {
-  GObjectClass *gobject_class;
-  GstObjectClass *gstobject_class;
-
-  gobject_class = (GObjectClass *) klass;
-  gstobject_class = (GstObjectClass *) klass;
+  GObjectClass *gobject_class = (GObjectClass *) klass;
 
   parent_class = g_type_class_peek_parent (klass);
 
-  if (!g_thread_supported ())
-    g_thread_init (NULL);
-
-  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_bus_dispose);
-  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_bus_set_property);
-  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_bus_get_property);
+  gobject_class->dispose = gst_bus_dispose;
 
   /**
    * GstBus::sync-message:
@@ -235,9 +198,7 @@ gst_bus_init (GstBus * bus)
 static void
 gst_bus_dispose (GObject * object)
 {
-  GstBus *bus;
-
-  bus = GST_BUS (object);
+  GstBus *bus = GST_BUS (object);
 
   if (bus->queue) {
     GstMessage *message;
@@ -265,36 +226,6 @@ gst_bus_dispose (GObject * object)
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
-static void
-gst_bus_set_property (GObject * object, guint prop_id,
-    const GValue * value, GParamSpec * pspec)
-{
-  GstBus *bus;
-
-  bus = GST_BUS (object);
-
-  switch (prop_id) {
-    default:
-      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-      break;
-  }
-}
-
-static void
-gst_bus_get_property (GObject * object, guint prop_id,
-    GValue * value, GParamSpec * pspec)
-{
-  GstBus *bus;
-
-  bus = GST_BUS (object);
-
-  switch (prop_id) {
-    default:
-      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-      break;
-  }
-}
-
 static void
 gst_bus_wakeup_main_context (GstBus * bus)
 {
@@ -343,7 +274,7 @@ gst_bus_new (void)
 {
   GstBus *result;
 
-  result = g_object_new (gst_bus_get_type (), NULL);
+  result = g_object_newv (gst_bus_get_type (), 0, NULL);
   GST_DEBUG_OBJECT (result, "created new bus");
 
   return result;
@@ -472,7 +403,7 @@ is_flushing:
  * Check if there are pending messages on the bus that
  * should be handled.
  *
- * Returns: TRUE if there are messages on the bus to be handled, FALSE 
+ * Returns: TRUE if there are messages on the bus to be handled, FALSE
  * otherwise.
  *
  * MT safe.
@@ -1204,7 +1135,7 @@ gst_bus_sync_signal_handler (GstBus * bus, GstMessage * message, gpointer data)
  * Instructs GStreamer to emit the "sync-message" signal after running the bus's
  * sync handler. This function is here so that code can ensure that they can
  * synchronously receive messages without having to affect what the bin's sync
- * handler is. 
+ * handler is.
  *
  * This function may be called multiple times. To clean up, the caller is
  * responsible for calling gst_bus_disable_sync_message_emission() as many times
