@@ -29,10 +29,10 @@
  *
  * GhostPads are useful when organizing pipelines with #GstBin like elements.
  * The idea here is to create hierarchical element graphs. The bin element
- * contains a sub-graph. Now one would like to treat the bin-element like other
- * #GstElements. This is where GhostPads come into play. A GhostPad acts as a
- * proxy for another pad. Thus the bin can have sink and source ghost-pads that
- * are associated with sink and source pads of the child elements.
+ * contains a sub-graph. Now one would like to treat the bin-element like any
+ * other #GstElement. This is where GhostPads come into play. A GhostPad acts as
+ * a proxy for another pad. Thus the bin can have sink and source ghost-pads
+ * that are associated with sink and source pads of the child elements.
  *
  * If the target pad is known at creation time, gst_ghost_pad_new() is the
  * function to use to get a ghost-pad. Otherwise one can use gst_ghost_pad_new_no_target()
@@ -56,6 +56,7 @@
 #define GST_PROXY_PAD_PRIVATE(obj)      (GST_PROXY_PAD_CAST (obj)->priv)
 #define GST_PROXY_PAD_TARGET(pad)       (GST_PROXY_PAD_PRIVATE (pad)->target)
 #define GST_PROXY_PAD_INTERNAL(pad)     (GST_PROXY_PAD_PRIVATE (pad)->internal)
+#define GST_PROXY_PAD_RETARGET(pad)     (GST_PROXY_PAD_PRIVATE (pad)->retarget)
 #define GST_PROXY_GET_LOCK(pad) (GST_PROXY_PAD_PRIVATE (pad)->proxy_lock)
 #define GST_PROXY_LOCK(pad)     (g_mutex_lock (GST_PROXY_GET_LOCK (pad)))
 #define GST_PROXY_UNLOCK(pad)   (g_mutex_unlock (GST_PROXY_GET_LOCK (pad)))
@@ -66,6 +67,7 @@ struct _GstProxyPadPrivate
   GMutex *proxy_lock;
   GstPad *target;
   GstPad *internal;
+  gboolean retarget;
 };
 
 G_DEFINE_TYPE (GstProxyPad, gst_proxy_pad, GST_TYPE_PAD);
@@ -81,28 +83,9 @@ static xmlNodePtr gst_proxy_pad_save_thyself (GstObject * object,
 #endif
 
 static void on_src_target_notify (GstPad * target,
-    GParamSpec * unused, GstGhostPad * pad);
+    GParamSpec * unused, gpointer user_data);
 
 
-static void
-gst_proxy_pad_class_init (GstProxyPadClass * klass)
-{
-  GObjectClass *gobject_class = (GObjectClass *) klass;
-
-  g_type_class_add_private (klass, sizeof (GstProxyPadPrivate));
-
-  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_proxy_pad_dispose);
-  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_proxy_pad_finalize);
-
-#ifndef GST_DISABLE_LOADSAVE
-  {
-    GstObjectClass *gstobject_class = (GstObjectClass *) klass;
-
-    gstobject_class->save_thyself =
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_save_thyself);
-  }
-#endif
-}
 
 static const GstQueryType *
 gst_proxy_pad_do_query_type (GstPad * pad)
@@ -142,20 +125,6 @@ gst_proxy_pad_do_query (GstPad * pad, GstQuery * query)
   return res;
 }
 
-static GList *
-gst_proxy_pad_do_internal_link (GstPad * pad)
-{
-  GList *res = NULL;
-  GstPad *target = gst_proxy_pad_get_target (pad);
-
-  if (target) {
-    res = gst_pad_get_internal_links (target);
-    gst_object_unref (target);
-  }
-
-  return res;
-}
-
 static GstIterator *
 gst_proxy_pad_do_iterate_internal_links (GstPad * pad)
 {
@@ -193,6 +162,17 @@ gst_proxy_pad_do_chain (GstPad * pad, GstBuffer * buffer)
   return res;
 }
 
+static GstFlowReturn
+gst_proxy_pad_do_chain_list (GstPad * pad, GstBufferList * list)
+{
+  GstFlowReturn res;
+  GstPad *internal = GST_PROXY_PAD_INTERNAL (pad);
+
+  res = gst_pad_push_list (internal, list);
+
+  return res;
+}
+
 static GstFlowReturn
 gst_proxy_pad_do_getrange (GstPad * pad, guint64 offset, guint size,
     GstBuffer ** buffer)
@@ -225,10 +205,12 @@ gst_proxy_pad_do_getcaps (GstPad * pad)
 
   if (target) {
     /* if we have a real target, proxy the call */
-    res = gst_pad_get_caps (target);
-    gst_object_unref (target);
+    res = gst_pad_get_caps_reffed (target);
 
-    GST_DEBUG_OBJECT (pad, "get caps of target: %" GST_PTR_FORMAT, res);
+    GST_DEBUG_OBJECT (pad, "get caps of target %s:%s : %" GST_PTR_FORMAT,
+        GST_DEBUG_PAD_NAME (target), res);
+
+    gst_object_unref (target);
 
     /* filter against the template */
     if (templ && res) {
@@ -371,6 +353,59 @@ gst_proxy_pad_get_target (GstPad * pad)
   return target;
 }
 
+static void
+gst_proxy_pad_do_unlink (GstPad * pad)
+{
+  GstPad *internal;
+
+  /* don't do anything if this unlink resulted from retargeting the pad
+   * controlled by the ghostpad. We only want to invalidate the target pad when
+   * the element suddently unlinked with our internal pad. */
+  if (GST_PROXY_PAD_RETARGET (pad))
+    return;
+
+  internal = GST_PROXY_PAD_INTERNAL (pad);
+
+  GST_DEBUG_OBJECT (pad, "pad is unlinked");
+
+  gst_proxy_pad_set_target (internal, NULL);
+}
+
+static void
+gst_proxy_pad_class_init (GstProxyPadClass * klass)
+{
+  GObjectClass *gobject_class = (GObjectClass *) klass;
+
+  g_type_class_add_private (klass, sizeof (GstProxyPadPrivate));
+
+  gobject_class->dispose = gst_proxy_pad_dispose;
+  gobject_class->finalize = gst_proxy_pad_finalize;
+
+#ifndef GST_DISABLE_LOADSAVE
+  {
+    GstObjectClass *gstobject_class = (GstObjectClass *) klass;
+
+    gstobject_class->save_thyself =
+        GST_DEBUG_FUNCPTR (gst_proxy_pad_save_thyself);
+  }
+#endif
+  /* Register common function pointer descriptions */
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_query_type);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_event);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_query);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_iterate_internal_links);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_getcaps);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_acceptcaps);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_fixatecaps);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_setcaps);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_unlink);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_bufferalloc);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_chain);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_chain_list);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_getrange);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_proxy_pad_do_checkgetrange);
+}
+
 static void
 gst_proxy_pad_dispose (GObject * object)
 {
@@ -409,23 +444,17 @@ gst_proxy_pad_init (GstProxyPad * ppad)
       GST_TYPE_PROXY_PAD, GstProxyPadPrivate);
   GST_PROXY_GET_LOCK (pad) = g_mutex_new ();
 
-  gst_pad_set_query_type_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_query_type));
-  gst_pad_set_event_function (pad, GST_DEBUG_FUNCPTR (gst_proxy_pad_do_event));
-  gst_pad_set_query_function (pad, GST_DEBUG_FUNCPTR (gst_proxy_pad_do_query));
-  gst_pad_set_internal_link_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_internal_link));
+  gst_pad_set_query_type_function (pad, gst_proxy_pad_do_query_type);
+  gst_pad_set_event_function (pad, gst_proxy_pad_do_event);
+  gst_pad_set_query_function (pad, gst_proxy_pad_do_query);
   gst_pad_set_iterate_internal_links_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_iterate_internal_links));
+      gst_proxy_pad_do_iterate_internal_links);
 
-  gst_pad_set_getcaps_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_getcaps));
-  gst_pad_set_acceptcaps_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_acceptcaps));
-  gst_pad_set_fixatecaps_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_fixatecaps));
-  gst_pad_set_setcaps_function (pad,
-      GST_DEBUG_FUNCPTR (gst_proxy_pad_do_setcaps));
+  gst_pad_set_getcaps_function (pad, gst_proxy_pad_do_getcaps);
+  gst_pad_set_acceptcaps_function (pad, gst_proxy_pad_do_acceptcaps);
+  gst_pad_set_fixatecaps_function (pad, gst_proxy_pad_do_fixatecaps);
+  gst_pad_set_setcaps_function (pad, gst_proxy_pad_do_setcaps);
+  gst_pad_set_unlink_function (pad, gst_proxy_pad_do_unlink);
 }
 
 #ifndef GST_DISABLE_LOADSAVE
@@ -502,16 +531,6 @@ G_DEFINE_TYPE (GstGhostPad, gst_ghost_pad, GST_TYPE_PROXY_PAD);
 
 static void gst_ghost_pad_dispose (GObject * object);
 
-static void
-gst_ghost_pad_class_init (GstGhostPadClass * klass)
-{
-  GObjectClass *gobject_class = (GObjectClass *) klass;
-
-  g_type_class_add_private (klass, sizeof (GstGhostPadPrivate));
-
-  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_ghost_pad_dispose);
-}
-
 /* see gstghostpad design docs */
 static gboolean
 gst_ghost_pad_internal_do_activate_push (GstPad * pad, gboolean active)
@@ -652,56 +671,87 @@ link_failed:
 static void
 gst_ghost_pad_do_unlink (GstPad * pad)
 {
-  GstPad *target;
   GstPad *internal;
 
-  target = gst_proxy_pad_get_target (pad);
   internal = GST_PROXY_PAD_INTERNAL (pad);
 
   GST_DEBUG_OBJECT (pad, "unlinking ghostpad");
 
   /* The target of the internal pad is no longer valid */
   gst_proxy_pad_set_target (internal, NULL);
-
-  if (target) {
-    if (GST_PAD_UNLINKFUNC (target))
-      GST_PAD_UNLINKFUNC (target) (target);
-
-    gst_object_unref (target);
-  }
 }
 
 static void
 on_int_notify (GstPad * internal, GParamSpec * unused, GstGhostPad * pad)
 {
   GstCaps *caps;
+  gboolean changed;
 
   g_object_get (internal, "caps", &caps, NULL);
 
+  GST_DEBUG_OBJECT (pad, "notified %p %" GST_PTR_FORMAT, caps, caps);
+
   GST_OBJECT_LOCK (pad);
-  gst_caps_replace (&(GST_PAD_CAPS (pad)), caps);
+  changed = (GST_PAD_CAPS (pad) != caps);
+  if (changed)
+    gst_caps_replace (&(GST_PAD_CAPS (pad)), caps);
   GST_OBJECT_UNLOCK (pad);
 
-  g_object_notify (G_OBJECT (pad), "caps");
+  if (changed)
+    g_object_notify (G_OBJECT (pad), "caps");
+
   if (caps)
     gst_caps_unref (caps);
 }
 
 static void
-on_src_target_notify (GstPad * target, GParamSpec * unused, GstGhostPad * pad)
+on_src_target_notify (GstPad * target, GParamSpec * unused, gpointer user_data)
 {
+  GstProxyPad *proxypad;
+  GstGhostPad *gpad;
   GstCaps *caps;
+  gboolean changed;
 
   g_object_get (target, "caps", &caps, NULL);
 
-  GST_OBJECT_LOCK (pad);
-  gst_caps_replace (&(GST_PAD_CAPS (pad)), caps);
-  GST_OBJECT_UNLOCK (pad);
+  GST_OBJECT_LOCK (target);
+  /* First check if the peer is still available and our proxy pad */
+  if (!GST_PAD_PEER (target) || !GST_IS_PROXY_PAD (GST_PAD_PEER (target))) {
+    GST_OBJECT_UNLOCK (target);
+    goto done;
+  }
 
-  g_object_notify (G_OBJECT (pad), "caps");
+  proxypad = GST_PROXY_PAD (GST_PAD_PEER (target));
+  GST_PROXY_LOCK (proxypad);
+  /* Now check if the proxypad's internal pad is still there and
+   * a ghostpad */
+  if (!GST_PROXY_PAD_INTERNAL (proxypad) ||
+      !GST_IS_GHOST_PAD (GST_PROXY_PAD_INTERNAL (proxypad))) {
+    GST_OBJECT_UNLOCK (target);
+    GST_PROXY_UNLOCK (proxypad);
+    goto done;
+  }
+  gpad = GST_GHOST_PAD (GST_PROXY_PAD_INTERNAL (proxypad));
+  g_object_ref (gpad);
+  GST_PROXY_UNLOCK (proxypad);
+  GST_OBJECT_UNLOCK (target);
+
+  GST_OBJECT_LOCK (gpad);
+
+  GST_DEBUG_OBJECT (gpad, "notified %p %" GST_PTR_FORMAT, caps, caps);
+
+  changed = (GST_PAD_CAPS (gpad) != caps);
+  if (changed)
+    gst_caps_replace (&(GST_PAD_CAPS (gpad)), caps);
+  GST_OBJECT_UNLOCK (gpad);
+  if (changed)
+    g_object_notify (G_OBJECT (gpad), "caps");
+
+  g_object_unref (gpad);
+
+done:
   if (caps)
     gst_caps_unref (caps);
-
 }
 
 static gboolean
@@ -713,18 +763,49 @@ gst_ghost_pad_do_setcaps (GstPad * pad, GstCaps * caps)
   return gst_proxy_pad_do_setcaps (pad, caps);
 }
 
+static GstIterator *
+gst_ghost_pad_do_iterate_internal_links (GstPad * pad)
+{
+  GstIterator *res = NULL;
+  GstPad *internal = GST_PROXY_PAD_INTERNAL (GST_GHOST_PAD_CAST (pad));
+
+  if (internal) {
+    res =
+        gst_iterator_new_single (GST_TYPE_PAD, internal,
+        (GstCopyFunction) gst_object_ref, (GFreeFunc) gst_object_unref);
+  }
+
+  return res;
+}
+
+static void
+gst_ghost_pad_class_init (GstGhostPadClass * klass)
+{
+  GObjectClass *gobject_class = (GObjectClass *) klass;
+
+  g_type_class_add_private (klass, sizeof (GstGhostPadPrivate));
+
+  gobject_class->dispose = gst_ghost_pad_dispose;
+
+  GST_DEBUG_REGISTER_FUNCPTR (gst_ghost_pad_do_setcaps);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_ghost_pad_do_activate_pull);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_ghost_pad_do_activate_push);
+  GST_DEBUG_REGISTER_FUNCPTR (gst_ghost_pad_do_link);
+}
+
 static void
 gst_ghost_pad_init (GstGhostPad * pad)
 {
   GST_GHOST_PAD_PRIVATE (pad) = G_TYPE_INSTANCE_GET_PRIVATE (pad,
       GST_TYPE_GHOST_PAD, GstGhostPadPrivate);
 
-  gst_pad_set_setcaps_function (GST_PAD_CAST (pad),
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_do_setcaps));
+  gst_pad_set_setcaps_function (GST_PAD_CAST (pad), gst_ghost_pad_do_setcaps);
   gst_pad_set_activatepull_function (GST_PAD_CAST (pad),
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_do_activate_pull));
+      gst_ghost_pad_do_activate_pull);
   gst_pad_set_activatepush_function (GST_PAD_CAST (pad),
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_do_activate_push));
+      gst_ghost_pad_do_activate_push);
+  gst_pad_set_iterate_internal_links_function (GST_PAD_CAST (pad),
+      gst_ghost_pad_do_iterate_internal_links);
 }
 
 static void
@@ -805,22 +886,17 @@ gst_ghost_pad_construct (GstGhostPad * gpad)
 
   /* Set directional padfunctions for ghostpad */
   if (dir == GST_PAD_SINK) {
-    gst_pad_set_bufferalloc_function (pad,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_bufferalloc));
-    gst_pad_set_chain_function (pad,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_chain));
+    gst_pad_set_bufferalloc_function (pad, gst_proxy_pad_do_bufferalloc);
+    gst_pad_set_chain_function (pad, gst_proxy_pad_do_chain);
+    gst_pad_set_chain_list_function (pad, gst_proxy_pad_do_chain_list);
   } else {
-    gst_pad_set_getrange_function (pad,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_getrange));
-    gst_pad_set_checkgetrange_function (pad,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_checkgetrange));
+    gst_pad_set_getrange_function (pad, gst_proxy_pad_do_getrange);
+    gst_pad_set_checkgetrange_function (pad, gst_proxy_pad_do_checkgetrange);
   }
 
   /* link/unlink functions */
-  gst_pad_set_link_function (pad, GST_DEBUG_FUNCPTR (gst_ghost_pad_do_link));
-  gst_pad_set_unlink_function (pad,
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_do_unlink));
-
+  gst_pad_set_link_function (pad, gst_ghost_pad_do_link);
+  gst_pad_set_unlink_function (pad, gst_ghost_pad_do_unlink);
 
   /* INTERNAL PAD, it always exists and is child of the ghostpad */
   otherdir = (dir == GST_PAD_SRC) ? GST_PAD_SINK : GST_PAD_SRC;
@@ -839,15 +915,13 @@ gst_ghost_pad_construct (GstGhostPad * gpad)
 
   /* Set directional padfunctions for internal pad */
   if (dir == GST_PAD_SRC) {
-    gst_pad_set_bufferalloc_function (internal,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_bufferalloc));
-    gst_pad_set_chain_function (internal,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_chain));
+    gst_pad_set_bufferalloc_function (internal, gst_proxy_pad_do_bufferalloc);
+    gst_pad_set_chain_function (internal, gst_proxy_pad_do_chain);
+    gst_pad_set_chain_list_function (internal, gst_proxy_pad_do_chain_list);
   } else {
-    gst_pad_set_getrange_function (internal,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_getrange));
+    gst_pad_set_getrange_function (internal, gst_proxy_pad_do_getrange);
     gst_pad_set_checkgetrange_function (internal,
-        GST_DEBUG_FUNCPTR (gst_proxy_pad_do_checkgetrange));
+        gst_proxy_pad_do_checkgetrange);
   }
 
   GST_PROXY_LOCK (pad);
@@ -862,7 +936,7 @@ gst_ghost_pad_construct (GstGhostPad * gpad)
    * At this point, the GstGhostPad has a refcount of 1, and the internal pad has
    * a refcount of 1.
    * When the refcount of the GstGhostPad drops to 0, the ghostpad will dispose
-   * it's refcount on the internal pad in the dispose method by un-parenting it.
+   * its refcount on the internal pad in the dispose method by un-parenting it.
    * This is why we don't take extra refcounts in the assignments below
    */
   GST_PROXY_PAD_INTERNAL (pad) = internal;
@@ -879,9 +953,9 @@ gst_ghost_pad_construct (GstGhostPad * gpad)
 
   /* special activation functions for the internal pad */
   gst_pad_set_activatepull_function (internal,
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_internal_do_activate_pull));
+      gst_ghost_pad_internal_do_activate_pull);
   gst_pad_set_activatepush_function (internal,
-      GST_DEBUG_FUNCPTR (gst_ghost_pad_internal_do_activate_push));
+      gst_ghost_pad_internal_do_activate_push);
 
   GST_PROXY_UNLOCK (pad);
 
@@ -921,7 +995,7 @@ gst_ghost_pad_new_full (const gchar * name, GstPadDirection dir,
   if (!gst_ghost_pad_construct (ret))
     goto construct_failed;
 
-  return GST_PAD (ret);
+  return GST_PAD_CAST (ret);
 
 construct_failed:
   /* already logged */
@@ -1070,9 +1144,9 @@ gst_ghost_pad_new_no_target_from_template (const gchar * name,
 
 /**
  * gst_ghost_pad_get_target:
- * @gpad: the #GstGhostpad
+ * @gpad: the #GstGhostPad
  *
- * Get the target pad of #gpad. Unref target pad after usage.
+ * Get the target pad of @gpad. Unref target pad after usage.
  *
  * Returns: the target #GstPad, can be NULL if the ghostpad
  * has no target set. Unref target pad after usage.
@@ -1093,7 +1167,7 @@ gst_ghost_pad_get_target (GstGhostPad * gpad)
 
 /**
  * gst_ghost_pad_set_target:
- * @gpad: the #GstGhostpad
+ * @gpad: the #GstGhostPad
  * @newtarget: the new pad target
  *
  * Set the new target of the ghostpad @gpad. Any existing target
@@ -1112,8 +1186,11 @@ gst_ghost_pad_set_target (GstGhostPad * gpad, GstPad * newtarget)
   GstPadLinkReturn lret;
 
   g_return_val_if_fail (GST_IS_GHOST_PAD (gpad), FALSE);
+  g_return_val_if_fail (GST_PAD_CAST (gpad) != newtarget, FALSE);
+  g_return_val_if_fail (newtarget != GST_PROXY_PAD_INTERNAL (gpad), FALSE);
 
-  GST_PROXY_LOCK (gpad);
+  /* no need for locking, the internal pad's lifecycle is directly linked to the
+   * ghostpad's */
   internal = GST_PROXY_PAD_INTERNAL (gpad);
 
   if (newtarget)
@@ -1122,27 +1199,31 @@ gst_ghost_pad_set_target (GstGhostPad * gpad, GstPad * newtarget)
     GST_DEBUG_OBJECT (gpad, "clearing target");
 
   /* clear old target */
+  GST_PROXY_LOCK (gpad);
   if ((oldtarget = GST_PROXY_PAD_TARGET (gpad))) {
     if (GST_PAD_IS_SRC (oldtarget)) {
       g_signal_handlers_disconnect_by_func (oldtarget,
-          on_src_target_notify, gpad);
+          (gpointer) on_src_target_notify, NULL);
     }
 
-    /* if we have an internal pad, unlink */
-    if (internal) {
-      if (GST_PAD_IS_SRC (internal))
-        gst_pad_unlink (internal, oldtarget);
-      else
-        gst_pad_unlink (oldtarget, internal);
-    }
+    GST_PROXY_PAD_RETARGET (internal) = TRUE;
+
+    /* unlink internal pad */
+    if (GST_PAD_IS_SRC (internal))
+      gst_pad_unlink (internal, oldtarget);
+    else
+      gst_pad_unlink (oldtarget, internal);
+
+    GST_PROXY_PAD_RETARGET (internal) = FALSE;
   }
 
   result = gst_proxy_pad_set_target_unlocked (GST_PAD_CAST (gpad), newtarget);
+  GST_PROXY_UNLOCK (gpad);
 
   if (result && newtarget) {
     if (GST_PAD_IS_SRC (newtarget)) {
       g_signal_connect (newtarget, "notify::caps",
-          G_CALLBACK (on_src_target_notify), gpad);
+          G_CALLBACK (on_src_target_notify), NULL);
     }
 
     /* and link to internal pad */
@@ -1156,7 +1237,6 @@ gst_ghost_pad_set_target (GstGhostPad * gpad, GstPad * newtarget)
     if (lret != GST_PAD_LINK_OK)
       goto link_failed;
   }
-  GST_PROXY_UNLOCK (gpad);
 
   return result;
 
@@ -1166,6 +1246,7 @@ link_failed:
     GST_WARNING_OBJECT (gpad, "could not link internal and target, reason:%d",
         lret);
     /* and unset target again */
+    GST_PROXY_LOCK (gpad);
     gst_proxy_pad_set_target_unlocked (GST_PAD_CAST (gpad), NULL);
     GST_PROXY_UNLOCK (gpad);
     return FALSE;
