@@ -39,6 +39,7 @@
 /* for constructing an entry name */
 #include "gstelement.h"
 #include "gstpad.h"
+#include "gstinfo.h"
 
 /* Index signals and args */
 enum
@@ -54,6 +55,9 @@ enum
       /* FILL ME */
 };
 
+GST_DEBUG_CATEGORY_STATIC (index_debug);
+#define GST_CAT_DEFAULT index_debug
+
 static void gst_index_class_init (GstIndexClass * klass);
 static void gst_index_init (GstIndex * index);
 static void gst_index_finalize (GObject * object);
@@ -94,12 +98,9 @@ gst_index_resolver_get_type (void)
 {
   static GType index_resolver_type = 0;
   static const GEnumValue index_resolver[] = {
-    {GST_INDEX_RESOLVER_CUSTOM, "GST_INDEX_RESOLVER_CUSTOM",
-        "Use a custom resolver"},
-    {GST_INDEX_RESOLVER_GTYPE, "GST_INDEX_RESOLVER_GTYPE",
-        "Resolve an object to its GType[.padname]"},
-    {GST_INDEX_RESOLVER_PATH, "GST_INDEX_RESOLVER_PATH",
-        "Resolve an object to its path in the pipeline"},
+    {GST_INDEX_RESOLVER_CUSTOM, "GST_INDEX_RESOLVER_CUSTOM", "custom"},
+    {GST_INDEX_RESOLVER_GTYPE, "GST_INDEX_RESOLVER_GTYPE", "gtype"},
+    {GST_INDEX_RESOLVER_PATH, "GST_INDEX_RESOLVER_PATH", "path"},
     {0, NULL, NULL},
   };
 
@@ -123,38 +124,18 @@ gst_index_entry_get_type (void)
   return index_entry_type;
 }
 
-
-GType
-gst_index_get_type (void)
-{
-  static GType index_type = 0;
-
-  if (!index_type) {
-    static const GTypeInfo index_info = {
-      sizeof (GstIndexClass),
-      NULL,
-      NULL,
-      (GClassInitFunc) gst_index_class_init,
-      NULL,
-      NULL,
-      sizeof (GstIndex),
-      0,
-      (GInstanceInitFunc) gst_index_init,
-      NULL
-    };
-
-    index_type =
-        g_type_register_static (GST_TYPE_OBJECT, "GstIndex", &index_info, 0);
-  }
-  return index_type;
+#define _do_init \
+{ \
+  GST_DEBUG_CATEGORY_INIT (index_debug, "GST_INDEX", GST_DEBUG_BOLD, \
+      "Generic indexing support"); \
 }
 
+G_DEFINE_TYPE_WITH_CODE (GstIndex, gst_index, GST_TYPE_OBJECT, _do_init);
+
 static void
 gst_index_class_init (GstIndexClass * klass)
 {
-  GObjectClass *gobject_class;
-
-  gobject_class = G_OBJECT_CLASS (klass);
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
   parent_class = g_type_class_peek_parent (klass);
 
@@ -170,14 +151,15 @@ gst_index_class_init (GstIndexClass * klass)
       G_STRUCT_OFFSET (GstIndexClass, entry_added), NULL, NULL,
       gst_marshal_VOID__BOXED, G_TYPE_NONE, 1, GST_TYPE_INDEX_ENTRY);
 
-  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_index_set_property);
-  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_index_get_property);
-  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_index_finalize);
+  gobject_class->set_property = gst_index_set_property;
+  gobject_class->get_property = gst_index_get_property;
+  gobject_class->finalize = gst_index_finalize;
 
   g_object_class_install_property (gobject_class, ARG_RESOLVER,
       g_param_spec_enum ("resolver", "Resolver",
           "Select a predefined object to string mapper",
-          GST_TYPE_INDEX_RESOLVER, GST_INDEX_RESOLVER_PATH, G_PARAM_READWRITE));
+          GST_TYPE_INDEX_RESOLVER, GST_INDEX_RESOLVER_PATH,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 }
 
 static void
@@ -301,7 +283,7 @@ gst_index_new (void)
 {
   GstIndex *index;
 
-  index = g_object_new (gst_index_get_type (), NULL);
+  index = g_object_newv (gst_index_get_type (), 0, NULL);
 
   return index;
 }
@@ -634,6 +616,8 @@ static gboolean
 gst_index_gtype_resolver (GstIndex * index, GstObject * writer,
     gchar ** writer_string, gpointer data)
 {
+  g_return_val_if_fail (writer != NULL, FALSE);
+
   if (GST_IS_PAD (writer)) {
     GstElement *element =
         (GstElement *) gst_object_get_parent (GST_OBJECT (writer));
@@ -741,7 +725,7 @@ gst_index_add_entry (GstIndex * index, GstIndexEntry * entry)
     iclass->add_entry (index, entry);
   }
 
-  g_signal_emit (G_OBJECT (index), gst_index_signals[ENTRY_ADDED], 0, entry);
+  g_signal_emit (index, gst_index_signals[ENTRY_ADDED], 0, entry);
 }
 
 /**
